"""Open a Jupyter notebook"""

import click
import shutil
from pathlib import Path, PurePath
from clease_gui import get_assets_path
from .main_cli import clease_gui_cli

__all__ = []

DEFAULT_FILENAME = 'clease_gui.ipynb'


def copy_default_notebook(filename: PurePath) -> None:
    src = get_assets_path() / 'default_notebook.ipynb'
    if filename.exists():
        raise RuntimeError(f'File {filename} already exists.')
    shutil.copy(src, filename)


@clease_gui_cli.command()
@click.option('-f',
              '--filename',
              type=str,
              default=DEFAULT_FILENAME,
              required=False,
              help=f"""
Create a mew notebook to be used with the CLEASE GUI.
Defaults to "{DEFAULT_FILENAME}".""")
def new(filename: str) -> None:
    """Create a mew notebook to be used with the CLEASE GUI.
    """
    file_path = Path(filename)
    # Ensure we're making an .ipynb file
    file_path = file_path.with_suffix('.ipynb')
    if file_path.exists():
        click.echo(f'File "{file_path}" already exists.')
        return
    click.echo(f'Creating new notebook: {file_path}')
    copy_default_notebook(file_path)



import logging
from typing import Optional
from IPython.display import display, clear_output
import ipywidgets as widgets

from clease.settings import ClusterExpansionSettings

from clease_gui import utils, BaseDashboard, register_logger, update_statusbar

__all__ = ['ClusterDashboard']

logger = logging.getLogger(__name__)
register_logger(logger)


class ClusterDashboard(BaseDashboard):
    """Dashboard for printing and visualizing the available clusters/figures
    in the active settings object."""
    def initialize(self) -> None:
        # Output for displaying clusters table
        self.clusters_content_out = widgets.Output(layout=dict(
            overflow='auto',
            width='100%',
            height='350px',
        ))

        self.refresh_table_button = widgets.Button(
            description='Update Clusters Table')
        self.refresh_table_button.on_click(self._on_refresh_table_click)

        self.view_clusters_button = widgets.Button(description='View Clusters')
        self.view_clusters_button.on_click(self._on_view_clusters_click)

    def display(self) -> None:
        button_box = widgets.HBox(
            children=[self.refresh_table_button, self.view_clusters_button])
        display(button_box, self.clusters_content_out)

    @utils.disable_cls_widget('refresh_table_button')
    @update_statusbar
    def _on_refresh_table_click(self, b):
        try:
            logger.info('Updating cluster table.')
            self.refresh_clusters_table()
        except Exception as exc:
            self.log_error(logger, exc)

    def refresh_clusters_table(self):
        """Draw the clusters table in the clusters content outputs"""
        settings = self.get_settings()
        with self.clusters_content_out:
            clear_output(wait=True)

            if settings is None:
                print('No active settings available')
                return
            # This call will trigger calculating the cluster list if it doesn't exist
            n_clusters = len(self.settings.cluster_list)
            print(f'Number of clusters: {n_clusters:d}')
            print(settings.clusters_table())

    def get_settings(self) -> Optional[ClusterExpansionSettings]:
        """Get the active settings object."""
        return self.app_data.get(self.KEYS.SETTINGS, None)

    @utils.disable_cls_widget('view_clusters_button')
    @update_statusbar
    def _on_view_clusters_click(self, b):
        logger.info('Temporarily disabled.')
        return
        try:
            logger.info('Opening clusters in the ASE GUI')
            self.view_clusters()
        except Exception as exc:
            self.log_error(logger, exc)

    def view_clusters(self):
        """Visualize the clusters in the ASE GUI"""
        settings = self.get_settings()
        if settings is None:
            raise RuntimeError('No active settings available')
        settings.view_clusters()



# pylint: disable=invalid-name
from typing import NamedTuple
import re
from pathlib import Path

__all__ = ('__version__', 'version_info')

with Path(__file__).with_name('_version.txt').open('r') as f:
    __version__ = f.readline().strip()


class Version(NamedTuple):
    major: int
    minor: int
    micro: int
    tag: str = None

    @classmethod
    def from_str(cls, s: str) -> 'Version':
        m = re.match(r'(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-(\w*))?', s)
        major = m.group(1)
        minor = m.group(2)
        micro = m.group(3)
        tag = m.group(4)
        return cls(major, minor, micro, tag=tag)

    def __str__(self) -> str:
        """The full version as X.Y.Z-tag"""
        s = f'{self.major}.{self.minor}.{self.micro}'
        if self.tag is not None:
            s += f'-{self.tag}'
        return s

    def short_version(self) -> str:
        """The short X.Y version"""
        return f'{self.major}.{self.minor}'


version_info = Version.from_str(__version__)



"""This dashboard is for the plots"""
import logging
import matplotlib.pyplot as plt
from IPython.display import display, clear_output
import ipywidgets as widgets

import clease.plot_post_process
from clease_gui import register_logger, BaseDashboard, update_statusbar
import clease_gui as gui

__all__ = ['FitPlotDashboard']

logger = logging.getLogger(__name__)
register_logger(logger)


class FitPlotDashboard(BaseDashboard):
    def initialize(self):

        tabs = []
        titles = []

        plt.ioff()

        def add_tab(output, title):
            tabs.append(output)
            titles.append(title)
            with output:
                clear_output()

        self.fit_out = widgets.Output()
        add_tab(self.fit_out, 'Fit')

        self.residual_out = widgets.Output()
        add_tab(self.residual_out, 'Residuals')

        self.eci_out = widgets.Output()
        add_tab(self.eci_out, 'ECI')

        self.cv_out = widgets.Output()
        add_tab(self.cv_out, 'CV')

        self.convex_hull_out = widgets.Output()
        add_tab(self.convex_hull_out, 'Convex Hull')

        self.plot_btn = widgets.Button(description='Update Plots',
                                       button_style='primary')

        self.plot_btn.on_click(self._on_make_plot_click)

        self.tabs = widgets.Tab(children=tabs)
        for ii, title in enumerate(titles):
            self.tabs.set_title(ii, title)

        # Keep track of open figures, so we can close them again
        # if we overwrite them.
        self._figures = {}

    def display(self):
        display(self.plot_btn, self.tabs)

    @property
    def evaluator(self):
        try:
            return self.app_data[self.KEYS.EVALUATE]
        except KeyError:
            raise KeyError('No fit has been performed yet.')

    def _close_figure(self, name: str) -> None:
        """Close a figure object with a certain name
        if it exists"""
        if name in self._figures:
            logger.debug('Closing figure: %s', name)
            fig = self._figures[name]
            plt.close(fig=fig)
            # We no longer need to store this
            self._figures.pop(name)

    def _register_figure(self, name: str, fig) -> None:
        """Register a figure object with a name.
        Will close the old figure if it exists"""
        self._close_figure(name)
        self._figures[name] = fig

    def _make_convex_hull(self):
        evaluator = self.evaluator

        fig = clease.plot_post_process.plot_convex_hull(evaluator)

        with self.convex_hull_out:
            clear_output(wait=True)
            display(fig)
        self._register_figure('convex_hull', fig)

    def _make_residual_plot(self):
        evaluator = self.evaluator
        fig = clease.plot_post_process.plot_fit_residual(evaluator)
        btn = self._make_save_figure_button(fig, 'fit_residual.png')

        with self.residual_out:
            clear_output(wait=True)
            display(fig, btn)
        self._register_figure('residual', fig)

    def _make_cv_plot(self):
        evaluate = self.evaluator
        try:
            fig = clease.plot_post_process.plot_cv(evaluate)
            btn = self._make_save_figure_button(fig, 'fit_cv.png')
        except ValueError:
            # No CV scores
            fig = None
            btn = None

        with self.cv_out:
            clear_output(wait=True)
            if fig is not None:
                display(fig, btn)
                self._register_figure('cv_score', fig)
            else:
                print('No CV values, requires alpha optimization')

    def _make_fit_plot(self):
        evaluate = self.evaluator
        fig = clease.plot_post_process.plot_fit(evaluate)
        btn = self._make_save_figure_button(fig, 'fit_plot.png')

        with self.fit_out:
            clear_output(wait=True)
            display(fig, btn)
        self._register_figure('fit_plot', fig)

    def _make_eci_plot(self):
        evaluate = self.evaluator
        fig = clease.plot_post_process.plot_eci(evaluate)
        btn = self._make_save_figure_button(fig, 'fit_eci.png')

        with self.eci_out:
            clear_output(wait=True)
            display(fig, btn)
        self._register_figure('eci_plot', fig)

    def _make_save_figure_button(self, fig, filename):
        button = widgets.Button(description='Save Figure')

        def _on_click(b):
            cwd = self.app_data[self.KEYS.CWD]
            pa = cwd / filename
            fig.savefig(filename, dpi=400)
            logger.info('Saved figure to file: %s', str(pa))

        button.on_click(_on_click)
        return button

    @update_statusbar
    @gui.utils.disable_cls_widget('plot_btn')
    def _on_make_plot_click(self, b):
        try:
            self._make_fit_plot()
            self._make_residual_plot()
            self._make_eci_plot()
            self._make_cv_plot()
            self._make_convex_hull()
        except Exception as exc:
            logger.error(exc)

    @property
    def fit_results(self):
        return self.app_data.get('fit_results', None)



import logging
import uuid
from pathlib import Path
from IPython.display import display
import ipywidgets as widgets
import clease_gui as gui
import json

__all__ = ['MCRunDashboard']

logger = logging.getLogger(__name__)
gui.register_logger(logger)


class MCRunDashboard(gui.BaseDashboard):
    def initialize(self):
        s = "Export MC run data to JSON file for further analysis, "
        s += "or load MC run data from file."
        html = f"<h1 style='font-size:15px'> {s} </h1>"
        self.description = widgets.HTML(html)
        self.filename_widget = widgets.Text(description='Filename:',
                                            value='clease_mc.json',
                                            **self.DEFAULT_STYLE_KWARGS)

        self.mode_widget = widgets.Dropdown(
            description='MC mode:',
            options=[
                ('Canonical', 'canonical'),
            ],
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.save_button = widgets.Button(description='Export data')
        self.save_button.on_click(self._on_save_click)

        self.append_widget = widgets.Checkbox(
            description='Append data in save/load?',
            value=True,
            **self.DEFAULT_STYLE_KWARGS)

        self.load_data_button = widgets.Button(description='Load Data')
        self.load_data_button.on_click(self._on_load_click)

    def display(self):
        hbox = widgets.HBox(children=[self.save_button, self.load_data_button])
        display(
            self.description,
            self.mode_widget,
            self.filename_widget,
            self.append_widget,
            hbox,
        )

    def get_filename(self):
        cwd = self.get_cwd()
        return cwd / self.filename_widget.value

    @property
    def mc_mode(self):
        return self.mode_widget.value

    def get_data(self):
        """Retrieve the MC run data from the app data"""
        mode = self.mc_mode
        if mode == 'canonical':
            key = self.KEYS.CANONICAL_MC_DATA
        else:
            raise ValueError(f'Unknown mode: {mode}')

        try:
            return list(self.app_data[key])
        except KeyError:
            raise KeyError('No data found. Did you run the MC yet?')

    def set_data(self, data, append=False) -> None:
        """Set the data of the currently active MC mode. If append=False, it will
        override any existing runs, and otherwise it will only append runs which
        aren't already in the list, i.e. the UUID does not exist in the list."""
        # We shouldn't have UUID collisions, but just in case we made a mistake.
        assert len({get_uuid_from_run(run)
                    for run in data
                    }) == len(data), "Not all runs have unique ID's"
        mode = self.mc_mode
        if mode == 'canonical':
            key = self.KEYS.CANONICAL_MC_DATA
        else:
            raise ValueError(f'Unknown mode: {mode}')

        if append:
            all_data = self.app_data.get(key, [])
            # Iterate through incoming new data, only append the runs which are new
            uids = {get_uuid_from_run(run) for run in all_data}
            for run in data:
                new_id = get_uuid_from_run(run)
                if new_id in uids:
                    # We already have this run
                    continue
                all_data.append(run)
        else:
            # Just override all the data
            all_data = data
        self.app_data[key] = all_data
        logger.debug('Updated MC data for mode: %s', mode)

    def save_data(self, filename):
        data = self.get_data()
        data = self.append_data(data, filename)
        with open(filename, 'w') as file:
            json.dump(data, file, indent=4)

    @gui.update_statusbar
    @gui.utils.disable_cls_widget('save_button')
    def _on_save_click(self, b):
        try:
            fname = self.get_filename()
            logger.info('Exporting data to file: %s', fname)
            self.save_data(fname)
        except Exception as exc:
            self.log_error(logger, exc)

    def append_data(self, data, filename):
        if not self.append_widget.value:
            return data
        if not Path(filename).is_file():
            return data

        logger.info('Found existing data file. Appending current data.')
        # Read the data, and append ours
        with open(filename) as file:
            existing_data = json.load(file)

        if not isinstance(existing_data, list):
            raise TypeError((f'Loaded data from file "{filename}", '
                             'which was not of the expected type.'
                             f'Got {type(existing_data)}, expected list'))

        runs_to_save = set(get_uuid_from_run(run) for run in data)
        # figure out what data we already saved
        for run in existing_data:
            uid = get_uuid_from_run(run)
            # Remove the ID from the set if it exists
            # since we already have it
            runs_to_save.discard(uid)

        # Figure out runs to save
        data = [run for run in data if get_uuid_from_run(run) in runs_to_save]

        data = existing_data + data
        return data

    def load_data(self, filename):
        with open(filename) as file:
            data = json.load(file)

        self.set_data(data, append=self.append_widget.value)

    @gui.update_statusbar
    @gui.utils.disable_cls_widget('load_data_button')
    def _on_load_click(self, b):
        try:
            fname = self.get_filename()
            logger.info('Loading data from file: %s', fname)
            self.load_data(fname)
        except Exception as exc:
            self.log_error(logger, exc)
        logger.info('Data loaded succesfully.')


def get_uuid_from_run(run):
    meta = run.get('meta', {})
    uid = meta.get('run_id', None)
    if uid is not None:
        uid = uuid.UUID(uid)
    return uid



from .concentration import *
from .settings import *
from .cluster_dashboard import *
from .settings_maker import *

__all__ = (concentration.__all__ + settings.__all__ + settings_maker.__all__ +
           cluster_dashboard.__all__)



import logging
import functools
from contextlib import contextmanager
from ipywidgets import DOMWidget
from .logging_widget import register_logger

__all__ = [
    'is_value_change',
    'disable_widget_context',
    'disables_widget',
    'disable_cls_widget',
]

logger = logging.getLogger(__name__)
register_logger(logger)


def is_value_change(change: dict):
    """Determine if a widget change is due to a change in value.
    This change comes from the widget.observe() function.
    """
    return change['type'] == 'change' and change['name'] == 'value'


@contextmanager
def disable_widget_context(widget: DOMWidget):
    """Context manager which disables a widget.
    Usage:

    with disable_widget_context(my_widget):
        do_stuff()
    """
    if not isinstance(widget, DOMWidget):
        raise TypeError(
            f'Widget must be of type DOMWidget, got {type(widget)}')
    widget.disabled = True
    logger.debug('Widget disabled')
    try:
        yield
    finally:
        widget.disabled = False
        logger.debug('Widget enabled')


def disables_widget(widget):
    """Decorator which disables a widget while function is executing.
    Usage:
    @disables_widget(my_widget)
    def my_func():
        # While this function is executing, the widget "my_widget" is disabled.
        do_stuff()
    """
    def decorator(func):
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            with disable_widget_context(widget):
                return func(*args, **kwargs)

        return _wrapper

    return decorator


def disable_cls_widget(widget_name: str):
    """Decorator which disables a widget of a class which is a member of "self"
    Assumes the widget name exists in "self", i.e. it uses getattr(self, widget_name)
    to access the widget.

    @disables_widget('my_widget_name')
    def my_func(self):
        do_stuff()
    """
    def decorator(func):
        @functools.wraps(func)
        def _wrapper(self, *args, **kwargs):
            widget = getattr(self, widget_name)
            with disable_widget_context(widget):
                return func(self, *args, **kwargs)

        return _wrapper

    return decorator



from pathlib import Path
from clease import jsonio

__all__ = ['save_app_data', 'load_app_data', 'AppDataKeys']


class AppDataKeys:
    """Collection of keys which (may) be in the app_data.
    Keys starting with an '_' will not be saved in the app state. """
    # "private" keys
    CWD = '_cwd'
    STATUS = '_status'
    DEV_MODE = '_dev_mode'

    # Regular app data keys
    SUPERCELL = 'supercell'
    SETTINGS = 'settings'
    ECI = 'eci'
    CANONICAL_MC_DATA = 'canonical_mc_data'

    # The evaluator cannot be saved to file, so save it as private
    # instance of an Evaluate class
    EVALUATE = '_evaluator'


def save_app_data(app_data, fname):
    fname = Path(fname)
    data = app_data.copy()
    to_remove = []
    for key in data:
        # Find keys which we don't want to save
        # Any keys starting with a "_" we say
        # we don't want to save
        if key.startswith('_'):
            to_remove.append(key)
    for key in to_remove:
        data.pop(key, None)

    with fname.open('w') as file:
        jsonio.write_json(file, data)


def load_app_data(fname):
    fname = Path(fname)
    with fname.open() as file:
        return jsonio.read_json(file)



from setuptools import setup, find_packages

EXTRAS_REQUIRE = {
    'dev': (
        'pre-commit',
        'yapf',
        'prospector',
        'pylint',
        'twine',
        'build',
        'pyclean>=2.0.0',  # For cleaning __pycache__ and *.pyc
        'tox',
    ),
    'test': (
        'pytest>=4',
        'pytest-mock',
        'pytest-cases',
    ),
    'doc': ('sphinx', 'sphinx_rtd_theme'),
}

EXTRAS_REQUIRE['full'] = set(value for tup in EXTRAS_REQUIRE.values()
                             for value in tup)

setup(
    packages=find_packages(include=['clease_gui', 'clease_gui.*']),
    extras_require=EXTRAS_REQUIRE,
)



import logging
from IPython.display import display, clear_output
import ipywidgets as widgets
from clease.structgen import NewStructures
from clease_gui import register_logger, BaseDashboard, update_statusbar, utils
from . import StructureGenerationDashboard, GeneralNewStructSettingsDashboard

__all__ = ['NewStructureDashboard']

logger = logging.getLogger(__name__)
register_logger(logger)

# Register the clease structgen logger
logger_clease = logging.getLogger('clease.structgen.new_struct')
register_logger(logger_clease)


class NewStructureDashboard(BaseDashboard):
    """Main dashboard for generating structures"""
    def initialize(self):

        self.general_settings_out = widgets.Output()
        self.structure_gen_out = widgets.Output()

        self.general_new_struct_dashboard = GeneralNewStructSettingsDashboard(
            self.app_data)
        with self.general_settings_out:
            clear_output()
            self.general_new_struct_dashboard.display()

        self.struct_gen_dashboard = StructureGenerationDashboard(self.app_data)
        with self.structure_gen_out:
            clear_output()
            self.struct_gen_dashboard.display()

        self.tab = widgets.Tab(children=[
            self.general_settings_out,
            self.structure_gen_out,
        ])

        self.tab.set_title(0, 'General Settings')
        self.tab.set_title(1, 'Structure Generation')

        # Button for running the new structures generation
        self.generate_widget = widgets.Button(description='Generate',
                                              button_style='primary')
        self.generate_widget.on_click(self._on_generate_click)
        # Make a box which places the button centralized
        box_layout = widgets.Layout(display='flex',
                                    flex_flow='column',
                                    align_items='center',
                                    width='100%')
        self.generate_box = widgets.HBox(children=[self.generate_widget],
                                         layout=box_layout)

    def display(self):
        display(self.tab)
        display(self.generate_box)

    def _make_new_struct_instance(self):
        dash = self.general_new_struct_dashboard  # alias
        generation_number = dash.generation_number
        struct_per_gen = dash.struct_per_gen

        logger.debug('Generation number: %s, struct per gen: %s',
                     generation_number, struct_per_gen)

        settings = self.app_data.get(self.KEYS.SETTINGS, None)
        if settings is None:
            logger.error('Settings does not exist yet.')
            return None

        new_struct = NewStructures(
            settings,
            generation_number=generation_number,
            struct_per_gen=struct_per_gen,
        )
        return new_struct

    def get_generator_settings(self):
        """Collect the args and kwargs from the generation dashboard"""
        dash = self.struct_gen_dashboard  # Alias

        args = dash.get_args()
        kwargs = dash.get_kwargs()
        func_name = dash.get_generator_name()

        return dict(func_name=func_name, args=args, kwargs=kwargs)

    def _generate(self, new_struct, generator_settings):
        """Run the generate function on the new struct class"""
        func_name = generator_settings['func_name']
        args = generator_settings['args']
        kwargs = generator_settings['kwargs']

        # Execute the generation function
        logger.debug('Executing generation function: %s', func_name)
        logger.debug('Generator args: %s', args)
        logger.debug('Generator kwargs: %s', kwargs)
        getattr(new_struct, func_name)(*args, **kwargs)

    @update_statusbar
    def _on_generate_click(self, b):
        try:
            with utils.disable_widget_context(self.generate_widget):
                # Disable the generate more structures button
                self._do_generate()
        except Exception as exc:
            self.log_error(logger, exc)

    def _do_generate(self):
        logger.info('Generating new structures')
        generator_settings = self.get_generator_settings()
        logger.debug('Running with the following settings:\n%s',
                     generator_settings)
        new_struct = self._make_new_struct_instance()
        if new_struct is None:
            logger.error('Failed to make a NewStructures instance.')
            return
        self._generate(new_struct, generator_settings)

        logger.info('Structure generation complete!')
